@@ -14,7 +14,6 @@ from time import sleep
 from uuid import uuid4
 
 import IPython
-import boto3
 import botocore
 import filelock
 from IPython import get_ipython
@@ -26,6 +25,33 @@ from IPython.core.magic import (
 )
 from IPython.core.magic_arguments import magic_arguments, argument, parse_argstring
 import sparkmagic.utils.configuration as conf
+from ..resource.emr.auth import ClusterSessionCredentialsProvider
+from ..utils.boto_client_utils import get_boto3_session
+from ..utils.constants import (
+    LIBRARY_NAME,
+    SERVICE,
+    OPERATION,
+    SAGEMAKER_ANALYTICS_LOG_BASE_DIRECTORY,
+)
+from ..utils.service_metrics import (
+    records_service_metrics,
+)
+from ..utils.logging_utils import ServiceFileLogger
+
+from ..utils.emr_constants import (
+    IPYTHON_KERNEL,
+    MAGIC_KERNELS,
+    LANGUAGES_SUPPORTED,
+    AUTH_TYPE_SET,
+    KRB_FILE_PATH,
+    SPARK_SESSION_NAME_PREFIX,
+    INSTANCE_COLLECTION_TYPE_GROUP,
+    LIVY_DEFAULT_PORT,
+    AUTH_TYPE_NONE,
+    AUTH_TYPE_KERBEROS,
+    AUTH_TYPE_BASIC_ACCESS,
+    EMR_CONNECTION_LOG_FILE,
+)
 from ..utils.resource_check import check_host_and_port, is_ssl_enabled
 from ..utils.arn_util import RoleArnValidator
 
@@ -35,31 +61,13 @@ from ipywidgets import widgets
 from requests_kerberos import REQUIRED
 from sagemaker_studio_sparkmagic_lib.emr import EMRCluster
 from sagemaker_studio_sparkmagic_lib.kerberos import write_krb_conf
-from sparkmagic.utils.constants import LANG_SCALA, LANG_PYTHON
 
-IPYTHON_KERNEL = "IPythonKernel"
-PYSPARK_KERNEL = "PySparkKernel"
-SPARK_KERNEL = "SparkKernel"
-MAGIC_KERNELS = {PYSPARK_KERNEL, SPARK_KERNEL}
-SUCCESS = "SUCCESS"
-FAILURE = "FAILURE"
-LIBRARY_NAME = "sagemaker-analytics"
-LANGUAGES_SUPPORTED = {LANG_SCALA, LANG_PYTHON}
-SUPPORTED_SERVICES = {"emr"}
-SUPPORTED_OPERATIONS = {"connect"}
-AUTH_TYPE_SET = {"Kerberos", "None", "Basic_Access"}
-KRB_FILE_DIR = "/etc"
-KRB_FILE_PATH = os.path.join(KRB_FILE_DIR, "krb5.conf")
-SPARK_SESSION_NAME_PREFIX = "sagemaker_studio_analytics_spark_session"
-INSTANCE_COLLECTION_TYPE_GROUP = "INSTANCE_GROUP"
+# This logger emits logs to sys.stdout, which is visible in customer's notebook in Studio
+sys_out_logger = logging.getLogger(__name__)
+sys_out_logger.setLevel(logging.INFO)
+sys_out_logger.addHandler(logging.StreamHandler(sys.stdout))
 
-# Livy default port
-# https://docs.aws.amazon.com/emr/latest/ManagementGuide/emr-web-interfaces.html
-LIVY_DEFAULT_PORT = "8998"
-
-logger = logging.getLogger(__name__)
-logger.setLevel(logging.INFO)
-logger.addHandler(logging.StreamHandler(sys.stdout))
+sagemaker_analytics_service_logger = None
 
 
 def _magic_kernel_connect_to_emr(
@@ -194,13 +202,13 @@ def _get_lock():
     """
     lock = filelock.FileLock("{0}.lock".format(KRB_FILE_PATH))
     if lock.is_locked:
-        logger.debug("Lock is already acquired, waiting for 15s to check again")
+        sys_out_logger.debug("Lock is already acquired, waiting for 15s to check again")
         # Try to handle an unexpected kernel crash case
         sleep(15)
         # Check if lock is still locked
         if lock.is_locked:
             # Force release the lock as the kernel might have died silently
-            logger.debug(
+            sys_out_logger.debug(
                 "Lock is still acquired after wait, trying to force release lock before attempting "
                 "to acquire for processing"
             )
@@ -208,6 +216,7 @@ def _get_lock():
     return lock
 
 
+# IpythonWidgetBasedAuthProvider class must reside in this root file.
 class IpythonWidgetBasedAuthProvider:
     """
     Class responsible for retrieving credentials needed to connect to EMR cluster.
@@ -252,7 +261,7 @@ class IpythonWidgetBasedAuthProvider:
 
 
 def _connect_to_cluster(emr_client, args, emr_cluster, kernel_name, username, password):
-    if args.auth_type == "Basic_Access":
+    if args.auth_type == AUTH_TYPE_BASIC_ACCESS:
         _initiate_connect_based_on_kernel(
             emr_client=emr_client,
             args=args,
@@ -323,7 +332,7 @@ def _handle_kerberos_connectivity_failure(args, completed_process):
 
 
 def _get_magic_kernel_endpoint_magic_line(args, livy_endpoint, username, password):
-    if args.auth_type == "Basic_Access":
+    if args.auth_type == AUTH_TYPE_BASIC_ACCESS:
         return "-s {0} -t {1} -u {2} -p {3}".format(
             livy_endpoint, args.auth_type, username, password
         )
@@ -333,9 +342,10 @@ def _get_magic_kernel_endpoint_magic_line(args, livy_endpoint, username, passwor
 
 def _get_ipython_kernel_endpoint_magic_line(args, livy_endpoint, username, password):
     # Choice of uuid generator based on: https://docs.python.org/3/library/uuid.html
-    session_name = "{0}_{1}".format(SPARK_SESSION_NAME_PREFIX, uuid4().hex)
+    # Maintain same session name for sparkmagic session and livy session
+    session_name = conf.session_configs()["name"]
 
-    if args.auth_type == "Basic_Access":
+    if args.auth_type == AUTH_TYPE_BASIC_ACCESS:
         return "add -s {0} -l {1} -t {2} -u {3} -a {4} -p {5}".format(
             session_name,
             args.language,
@@ -494,14 +504,38 @@ def _initiate_emr_connect(args, boto_session, kernel_name):
 
     """
     Even if assumable role arn is None, we can pass it to the EMRCluster constructor and will be handled by
-    _get_boto3_session method call gracefully. So no need to do a explicit None check here.
+    get_boto3_session method call gracefully. So no need to do a explicit None check here.
     """
-    cluster = EMRCluster(cluster_id=args.cluster_id, role_arn=args.assumable_role_arn)
+    cluster = EMRCluster(
+        cluster_id=args.cluster_id, role_arn=args.assumable_role_arn, emr=emr_client
+    )
     _validate_cluster_auth_with_auth_type_provided(
         auth_type=args.auth_type, emr_cluster=cluster
     )
 
-    if cluster.is_krb_cluster or args.auth_type == "Basic_Access":
+    # If emr execution role is passed, the EMR IAM passthrough should be used and the token API is invoked
+    # to get user and password which are later used for basic auth.
+    if args.auth_type == AUTH_TYPE_BASIC_ACCESS and args.emr_execution_role_arn:
+        # Before EMR IAM Passthrough is GA, use emr-iam-passthrough-beta as service.
+        beta_emr_client = boto_session.client("emr-iam-passthrough-beta")
+
+        credentials = (
+            ClusterSessionCredentialsProvider().get_cluster_session_credentials(
+                beta_emr_client, args.cluster_id, args.emr_execution_role_arn
+            )
+        )
+        # wait for 3s for EMR token propagation
+        sleep(3)
+
+        _initiate_connect_based_on_kernel(
+            emr_client=emr_client,
+            args=args,
+            emr_cluster=cluster,
+            username=credentials.username,
+            password=credentials.password,
+            kernel_name=kernel_name,
+        )
+    elif cluster.is_krb_cluster or args.auth_type == AUTH_TYPE_BASIC_ACCESS:
         authProvider = IpythonWidgetBasedAuthProvider()
         authProvider.get_credentials_and_connect_to_cluster(
             emr_client=emr_client,
@@ -523,6 +557,13 @@ def _initiate_emr_connect(args, boto_session, kernel_name):
 def _initiate_connect_based_on_kernel(
     emr_client, args, emr_cluster, username, password, kernel_name
 ):
+
+    # Define livy session name. session_configs are passed to livy api.
+    session_name = f"{SPARK_SESSION_NAME_PREFIX}_{uuid4().hex}"
+    conf.override(conf.session_configs.__name__, {"name": session_name})
+
+    _set_logging_config()
+
     if kernel_name in MAGIC_KERNELS:
         _magic_kernel_connect_to_emr(
             emr_client=emr_client,
@@ -544,20 +585,31 @@ def _initiate_connect_based_on_kernel(
         )
 
 
+def _set_logging_config() -> None:
+    """
+    Sets `disable_existing_loggers` key on the dictconfig config used by sparkmagic to create loggers.
+    By default, loading a logger via `logging.config.dictConfig` disables all existing loggers, including our loggers.
+    python doc: https://docs.python.org/3/library/logging.config.html
+    sparkmagic ref: https://github.com/jupyter-incubator/sparkmagic/blob/master/hdijupyterutils/hdijupyterutils/log.py#L13
+    :return: None
+    """
+    logging_config = conf.logging_config()
+    logging_config["disable_existing_loggers"] = False
+    conf.override(conf.logging_config.__name__, logging_config)
+
+
 def _validate_cluster_auth_with_auth_type_provided(auth_type, emr_cluster):
     is_cluster_kerberos_authenticated = emr_cluster.is_krb_cluster
     is_cluster_ldap_authenticated = _is_cluster_ldap(emr_cluster)
     is_cluster_no_auth = (
         not is_cluster_ldap_authenticated and not is_cluster_kerberos_authenticated
     )
-    # Check if auth type is kerberos or not
+
     if (
-        (auth_type == "None" and is_cluster_kerberos_authenticated)
-        or (auth_type == "Basic_Access" and is_cluster_kerberos_authenticated)
-        or (auth_type == "None" and is_cluster_ldap_authenticated)
-        or (auth_type == "Kerberos" and is_cluster_ldap_authenticated)
-        or (auth_type == "Basic_Access" and is_cluster_no_auth)
-        or (auth_type == "Kerberos" and is_cluster_no_auth)
+        (auth_type == AUTH_TYPE_NONE and is_cluster_kerberos_authenticated)
+        or (auth_type == AUTH_TYPE_NONE and is_cluster_ldap_authenticated)
+        or (auth_type == AUTH_TYPE_KERBEROS and is_cluster_ldap_authenticated)
+        or (auth_type == AUTH_TYPE_KERBEROS and is_cluster_no_auth)
     ):
         raise Exception(
             "Cluster auth type does not match provided auth {}".format(auth_type)
@@ -602,7 +654,7 @@ class SagemakerAnalytics(Magics):
         nargs="*",
         help="Command to execute. The command consists of a service name followed by a ' ' followed by an operation. "
         "Supported services are {0} and supported operations are {1}. For example a valid command is '{2}'.".format(
-            SUPPORTED_SERVICES, SUPPORTED_OPERATIONS, "emr connect"
+            SERVICE.list(), OPERATION.list(), "emr connect"
         ),
     )
     @argument(
@@ -635,6 +687,21 @@ class SagemakerAnalytics(Magics):
         help="The IAM role to assume when connecting to a cluster in a different AWS account. This argument is not "
         "required when connecting to a cluster in the same AWS account.",
     )
+    @argument(
+        "--session-tags",
+        type=str,
+        default=None,
+        help="The session tags to set while assuming the role(specified by assumable-role-arn). It's an optional "
+        "parameter passed in json format. Error will be thrown if no assumable-role-arn is specified. An "
+        "administrator must grant you the permissions necessary to pass session tags.",
+    )
+    @argument(
+        "--emr-execution-role-arn",
+        type=str,
+        default=None,
+        help="The IAM role passed to EMR to set up EMR job security context. This argument is optional and "
+        "used when IAM Passthrough feature is enabled for EMR.",
+    )
     def sm_analytics(self, line):
         usage = "Please look at usage of %sm_analytics by executing `%sm_analytics?`."
         user_input = line
@@ -645,18 +712,19 @@ class SagemakerAnalytics(Magics):
         service = args.command[0].lower()
         operation = args.command[1].lower()
 
+        global sagemaker_analytics_service_logger
+        kernel_name = None
+
         # emr
-        if service == "emr":
-            if operation == "connect":
+        if service == SERVICE.EMR:
+            if operation == OPERATION.CONNECT:
+                sagemaker_analytics_service_logger = ServiceFileLogger(
+                    EMR_CONNECTION_LOG_FILE,
+                    SAGEMAKER_ANALYTICS_LOG_BASE_DIRECTORY,
+                    sys_out_logger,
+                )
                 try:
                     kernel_name = type(IPython.Application.instance().kernel).__name__
-                    _validate_emr_args(args=args, usage=usage, kernel_name=kernel_name)
-
-                    # Only create boto session if absolutely needed
-                    boto_session = _get_boto3_session(args.assumable_role_arn)
-                    _initiate_emr_connect(
-                        args=args, boto_session=boto_session, kernel_name=kernel_name
-                    )
                 except Exception as e:
                     _echo_response_to_iopub_web_socket(
                         _build_response(
@@ -668,12 +736,42 @@ class SagemakerAnalytics(Magics):
                         )
                     )
                     raise e
+                _connect_to_emr(
+                    service=service,
+                    operation=operation,
+                    args=args,
+                    usage=usage,
+                    kernel_name=kernel_name,
+                    service_logger=sagemaker_analytics_service_logger,
+                )
             else:
                 raise Exception("Operation '{}' not found. {}".format(operation, usage))
         else:
             raise Exception("Service '{}' not found. {}".format(service, usage))
 
 
+@records_service_metrics
+def _connect_to_emr(service, operation, args, usage, kernel_name, service_logger):
+    try:
+        _validate_emr_args(args=args, usage=usage, kernel_name=kernel_name)
+        # Only create boto session if absolutely needed
+        boto_session = get_boto3_session(args.assumable_role_arn, args.session_tags)
+        _initiate_emr_connect(
+            args=args, boto_session=boto_session, kernel_name=kernel_name
+        )
+    except Exception as e:
+        _echo_response_to_iopub_web_socket(
+            _build_response(
+                cluster_id=args.cluster_id,
+                error_message=str(e),
+                success=False,
+                service=service,
+                operation=operation,
+            )
+        )
+        raise e
+
+
 # In order to actually use these magics, you must register them with a
 # running IPython.
 def load_ipython_extension(ipython):
@@ -687,31 +785,6 @@ def load_ipython_extension(ipython):
     ipython.register_magics(SagemakerAnalytics)
 
 
-def _get_boto3_session(role_arn):
-    """
-    Allows cross account cluster access
-    """
-    if not role_arn:
-        return boto3.session.Session()
-    else:
-        sts_client = boto3.client("sts")
-        try:
-            assume_role_object = sts_client.assume_role(
-                RoleArn=role_arn, RoleSessionName="SageMakerStudioUser"
-            )
-        except botocore.exceptions.ClientError as ce:
-            logger.debug(f"Failed to assume role: ({role_arn}) details. {ce.response}")
-            raise ValueError(
-                f"Unable to assume role(arn: {role_arn}). Ensure permissions are setup correctly"
-                f' Error: {ce.response["Error"]}'
-            ) from None
-        return boto3.Session(
-            aws_access_key_id=assume_role_object["Credentials"]["AccessKeyId"],
-            aws_secret_access_key=assume_role_object["Credentials"]["SecretAccessKey"],
-            aws_session_token=assume_role_object["Credentials"]["SessionToken"],
-        )
-
-
 def _list_instance_groups(emr_client, cluster_id):
     try:
         emr_instance_group_list = []
@@ -722,7 +795,7 @@ def _list_instance_groups(emr_client, cluster_id):
             emr_instance_group_list.extend(page["InstanceGroups"])
 
     except botocore.exceptions.ClientError as ce:
-        logger.error(
+        sys_out_logger.error(
             "Failed to list instance groups for EMR cluster({0}). {1}".format(
                 cluster_id, ce.response
             )
